'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var react = require('react');
var observerUtil = require('@nx-js/observer-util');

var tasks = new Set();
var batchCount = 0;

function add(task) {
  if (batchCount !== 0) {
    tasks.add(task);
  } else {
    runTask(task);
  }
}

function remove(task) {
  tasks.delete(task);
}

// must remove the task before executing it
// toherwise executed task caould start a new task list execution
// and recursively run itself infinite times
function runTask(task) {
  remove(task);
  task();
}

// this runs the passed function and delays all re-renders
// until the function is finished running
function batch(fn, args) {
  try {
    batchCount++;
    return fn.apply(this, args);
  } finally {
    batchCount--;
    if (batchCount === 0) {
      tasks.forEach(runTask);
    }
  }
}

// this creates and returns a batched version of the passed function
// the cache is necessary to always map the same thing to the same function
// which makes sure that addEventListener/removeEventListener pairs don't break
var cache = new Map();
function batchFn(fn) {
  if (typeof fn !== 'function') {
    return fn;
  }
  var batched = cache.get(fn);
  if (!batched) {
    batched = function () {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      return batch(fn, args);
    };
    cache.set(fn, batched);
  }
  return batched;
}

// batched window.addEventListener(cb) like callbacks
function batchCallbacks(functionWithCallbacks) {
  return function batchedCallbacks() {
    var args = [], len = arguments.length;
    while ( len-- ) args[ len ] = arguments[ len ];

    return functionWithCallbacks.apply(this, args.map(batchFn));
  };
}

// batches obj.onevent = fn like calls
function batchMethod(obj, method) {
  var descriptor = Object.getOwnPropertyDescriptor(obj, method);
  if (descriptor) {
    var newDescriptor = Object.assign({}, descriptor, {
      set: function set(value) {
        return descriptor.set.call(this, batchFn(value));
      }
    });
    Object.defineProperty(obj, method, newDescriptor);
  }
}

// try to find the global object
// it is window in the DOM and global in NodeJS and React Native
var globalObj;
if (typeof window !== 'undefined') {
  globalObj = window;
} else if (typeof global !== 'undefined') {
  globalObj = global;
}

// do a sync batching for the most common task sources
// this should be removed when React's own batching is improved in the future
if (globalObj) {
  globalObj.setTimeout = batchCallbacks(globalObj.setTimeout);
  globalObj.setInterval = batchCallbacks(globalObj.setInterval);
  if (globalObj.requestAnimationFrame) {
    globalObj.requestAnimationFrame = batchCallbacks(globalObj.requestAnimationFrame);
  }
  if (globalObj.requestIdleCallback) {
    globalObj.requestIdleCallback = batchCallbacks(globalObj.requestIdleCallback);
  }
  // eslint-disable-next-line
  Promise.prototype.then = batchCallbacks(Promise.prototype.then);
  // eslint-disable-next-line
  Promise.prototype.catch = batchCallbacks(Promise.prototype.catch);

  // this batches websocket event handlers
  if (globalObj.WebSocket) {
    var websocketHandlers = ['onopen', 'onmessage', 'onerror', 'onclose'];
    websocketHandlers.forEach(function (method) { return batchMethod(globalObj.WebSocket.prototype, method); });
  }
}

// batch addEventListener calls
if (globalObj.EventTarget) {
  EventTarget.prototype.addEventListener = batchCallbacks(EventTarget.prototype.addEventListener);
  EventTarget.prototype.removeEventListener = batchCallbacks(EventTarget.prototype.removeEventListener);
}

// HTTP event handlers are usually wrapped by Promises, which is covered above

var hasHooks = typeof react.useState === 'function';

var isInsideFunctionComponent = false;
var COMPONENT = Symbol('owner component');
var DUMMY_STATE = {};

function view(Comp) {
  var isStatelessComp = !(Comp.prototype && Comp.prototype.isReactComponent);

  var ReactiveComp;

  if (isStatelessComp && hasHooks) {
    // use a hook based reactive wrapper when we can
    ReactiveComp = react.memo(function (props) {
      var ref = react.useState();
      var setState = ref[1];

      // create a memoized reactive wrapper of the original component (render)
      // at the very first run of the component function
      var render = react.useMemo(function () {
        // run a dummy setState to schedule a new render
        var updater = function () { return setState(DUMMY_STATE); };

        return observerUtil.observe(Comp, {
          scheduler: {
            add: function () { return add(updater); },
            delete: function () { return remove(updater); }
          },
          lazy: true
        });
      }, []);

      // cleanup the reactive connections after the very last render of the component
      react.useEffect(function () {
        return function () { return observerUtil.unobserve(render); };
      }, []);

      // the isInsideFunctionComponent flag is used to toggle `store` behavior
      // based on where it was called from
      isInsideFunctionComponent = true;
      try {
        // run the reactive render instead of the original one
        return render(props);
      } finally {
        isInsideFunctionComponent = false;
      }
    });
  } else {
    var BaseComp = isStatelessComp ? react.Component : Comp;
    // a HOC which overwrites render, shouldComponentUpdate and componentWillUnmount
    // it decides when to run the new reactive methods and when to proxy to the original methods
    var ReactiveClassComp = (function (BaseComp) {
      function ReactiveClassComp(props, context) {
        var this$1 = this;

        BaseComp.call(this, props, context);

        this.state = this.state || {};
        this.state[COMPONENT] = this;

        // run a dummy setState to schedule a new render, avoid forceUpdate
        var updater = function () { return this$1.setState(DUMMY_STATE); };

        // create a reactive render for the component
        this.render = observerUtil.observe(this.render, {
          scheduler: {
            add: function () { return add(updater); },
            delete: function () { return remove(updater); }
          },
          lazy: true
        });
      }

      if ( BaseComp ) ReactiveClassComp.__proto__ = BaseComp;
      ReactiveClassComp.prototype = Object.create( BaseComp && BaseComp.prototype );
      ReactiveClassComp.prototype.constructor = ReactiveClassComp;

      ReactiveClassComp.prototype.render = function render () {
        return isStatelessComp ? Comp(this.props, this.context) : BaseComp.prototype.render.call(this);
      };

      // react should trigger updates on prop changes, while easyState handles store changes
      ReactiveClassComp.prototype.shouldComponentUpdate = function shouldComponentUpdate (nextProps, nextState) {
        var ref = this;
        var props = ref.props;
        var state = ref.state;

        // respect the case when user prohibits updates
        if (BaseComp.prototype.shouldComponentUpdate && !BaseComp.prototype.shouldComponentUpdate.call(this, nextProps, nextState)) {
          return false;
        }

        // return true if it is a reactive render or state changes
        if (state !== nextState) {
          return true;
        }

        // the component should update if any of its props shallowly changed value
        var keys = Object.keys(props);
        var nextKeys = Object.keys(nextProps);
        return nextKeys.length !== keys.length || nextKeys.some(function (key) { return props[key] !== nextProps[key]; });
      };

      // add a custom deriveStoresFromProps lifecyle method
      ReactiveClassComp.getDerivedStateFromProps = function getDerivedStateFromProps (props, state) {
        if (BaseComp.deriveStoresFromProps) {
          // inject all local stores and let the user mutate them directly
          var stores = mapStateToStores(state);
          BaseComp.deriveStoresFromProps.apply(this, [ props ].concat( stores ));
        }
        // respect user defined getDerivedStateFromProps
        if (BaseComp.getDerivedStateFromProps) {
          return BaseComp.getDerivedStateFromProps.call(this, props, state);
        }
        return null;
      };

      ReactiveClassComp.prototype.componentWillUnmount = function componentWillUnmount () {
        // call user defined componentWillUnmount
        if (BaseComp.prototype.componentWillUnmount) {
          BaseComp.prototype.componentWillUnmount.call(this);
        }
        // clean up memory used by Easy State
        observerUtil.unobserve(this.render);
      };

      return ReactiveClassComp;
    }(BaseComp));

    ReactiveComp = ReactiveClassComp;
  }

  ReactiveComp.displayName = Comp.displayName || Comp.name;
  // static props are inherited by class components,
  // but have to be copied for function components
  if (isStatelessComp) {
    for (var i = 0, list = Object.keys(Comp); i < list.length; i += 1) {
      var key = list[i];

      ReactiveComp[key] = Comp[key];
    }
  }

  return ReactiveComp;
}

function mapStateToStores(state) {
  // find store properties and map them to their none observable raw value
  // to do not trigger none static this.setState calls
  // from the static getDerivedStateFromProps lifecycle method
  var component = state[COMPONENT];
  return Object.keys(component).map(function (key) { return component[key]; }).filter(observerUtil.isObservable).map(observerUtil.raw);
}

function store(obj) {
  // do not create new versions of the store on every render
  // if it is a local store in a function component
  // create a memoized store at the first call instead
  if (hasHooks && isInsideFunctionComponent) {
    return react.useMemo(function () { return observerUtil.observable(obj); }, []);
  }
  return observerUtil.observable(obj);
}

exports.view = view;
exports.store = store;
exports.batch = batch;
