import { useState, Component, useEffect, useMemo, memo } from 'react';
import { observe, unobserve, raw, isObservable, observable } from '@nx-js/observer-util';

const tasks = new Set();
let batchCount = 0;

function add(task) {
  if (batchCount !== 0) {
    tasks.add(task);
  } else {
    runTask(task);
  }
}

function remove(task) {
  tasks.delete(task);
}

// must remove the task before executing it
// toherwise executed task caould start a new task list execution
// and recursively run itself infinite times
function runTask(task) {
  remove(task);
  task();
}

// this runs the passed function and delays all re-renders
// until the function is finished running
function batch(fn, args) {
  try {
    batchCount++;
    return fn.apply(this, args);
  } finally {
    batchCount--;
    if (batchCount === 0) {
      tasks.forEach(runTask);
    }
  }
}

// this creates and returns a batched version of the passed function
// the cache is necessary to always map the same thing to the same function
// which makes sure that addEventListener/removeEventListener pairs don't break
const cache = new Map();
function batchFn(fn) {
  if (typeof fn !== 'function') {
    return fn;
  }
  let batched = cache.get(fn);
  if (!batched) {
    batched = (...args) => batch(fn, args);
    cache.set(fn, batched);
  }
  return batched;
}

// batched window.addEventListener(cb) like callbacks
function batchCallbacks(functionWithCallbacks) {
  return function batchedCallbacks(...args) {
    return functionWithCallbacks.apply(this, args.map(batchFn));
  };
}

// batches obj.onevent = fn like calls
function batchMethod(obj, method) {
  const descriptor = Object.getOwnPropertyDescriptor(obj, method);
  if (descriptor) {
    const newDescriptor = Object.assign({}, descriptor, {
      set(value) {
        return descriptor.set.call(this, batchFn(value));
      }
    });
    Object.defineProperty(obj, method, newDescriptor);
  }
}

// try to find the global object
// it is window in the DOM and global in NodeJS and React Native
let globalObj;
if (typeof window !== 'undefined') {
  globalObj = window;
} else if (typeof global !== 'undefined') {
  globalObj = global;
}

// do a sync batching for the most common task sources
// this should be removed when React's own batching is improved in the future
if (globalObj) {
  globalObj.setTimeout = batchCallbacks(globalObj.setTimeout);
  globalObj.setInterval = batchCallbacks(globalObj.setInterval);
  if (globalObj.requestAnimationFrame) {
    globalObj.requestAnimationFrame = batchCallbacks(globalObj.requestAnimationFrame);
  }
  if (globalObj.requestIdleCallback) {
    globalObj.requestIdleCallback = batchCallbacks(globalObj.requestIdleCallback);
  }
  // eslint-disable-next-line
  Promise.prototype.then = batchCallbacks(Promise.prototype.then);
  // eslint-disable-next-line
  Promise.prototype.catch = batchCallbacks(Promise.prototype.catch);

  // this batches websocket event handlers
  if (globalObj.WebSocket) {
    const websocketHandlers = ['onopen', 'onmessage', 'onerror', 'onclose'];
    websocketHandlers.forEach(method => batchMethod(globalObj.WebSocket.prototype, method));
  }
}

// batch addEventListener calls
if (globalObj.EventTarget) {
  EventTarget.prototype.addEventListener = batchCallbacks(EventTarget.prototype.addEventListener);
  EventTarget.prototype.removeEventListener = batchCallbacks(EventTarget.prototype.removeEventListener);
}

// HTTP event handlers are usually wrapped by Promises, which is covered above

var hasHooks = typeof useState === 'function';

let isInsideFunctionComponent = false;
const COMPONENT = Symbol('owner component');
const DUMMY_STATE = {};

function view(Comp) {
  const isStatelessComp = !(Comp.prototype && Comp.prototype.isReactComponent);

  let ReactiveComp;

  if (isStatelessComp && hasHooks) {
    // use a hook based reactive wrapper when we can
    ReactiveComp = memo(props => {
      const [, setState] = useState();

      // create a memoized reactive wrapper of the original component (render)
      // at the very first run of the component function
      const render = useMemo(() => {
        // run a dummy setState to schedule a new render
        const updater = () => setState(DUMMY_STATE);

        return observe(Comp, {
          scheduler: {
            add: () => add(updater),
            delete: () => remove(updater)
          },
          lazy: true
        });
      }, []);

      // cleanup the reactive connections after the very last render of the component
      useEffect(() => {
        return () => unobserve(render);
      }, []);

      // the isInsideFunctionComponent flag is used to toggle `store` behavior
      // based on where it was called from
      isInsideFunctionComponent = true;
      try {
        // run the reactive render instead of the original one
        return render(props);
      } finally {
        isInsideFunctionComponent = false;
      }
    });
  } else {
    const BaseComp = isStatelessComp ? Component : Comp;
    // a HOC which overwrites render, shouldComponentUpdate and componentWillUnmount
    // it decides when to run the new reactive methods and when to proxy to the original methods
    class ReactiveClassComp extends BaseComp {
      constructor(props, context) {
        super(props, context);

        this.state = this.state || {};
        this.state[COMPONENT] = this;

        // run a dummy setState to schedule a new render, avoid forceUpdate
        const updater = () => this.setState(DUMMY_STATE);

        // create a reactive render for the component
        this.render = observe(this.render, {
          scheduler: {
            add: () => add(updater),
            delete: () => remove(updater)
          },
          lazy: true
        });
      }

      render() {
        return isStatelessComp ? Comp(this.props, this.context) : super.render();
      }

      // react should trigger updates on prop changes, while easyState handles store changes
      shouldComponentUpdate(nextProps, nextState) {
        const { props, state } = this;

        // respect the case when user prohibits updates
        if (super.shouldComponentUpdate && !super.shouldComponentUpdate(nextProps, nextState)) {
          return false;
        }

        // return true if it is a reactive render or state changes
        if (state !== nextState) {
          return true;
        }

        // the component should update if any of its props shallowly changed value
        const keys = Object.keys(props);
        const nextKeys = Object.keys(nextProps);
        return nextKeys.length !== keys.length || nextKeys.some(key => props[key] !== nextProps[key]);
      }

      // add a custom deriveStoresFromProps lifecyle method
      static getDerivedStateFromProps(props, state) {
        if (super.deriveStoresFromProps) {
          // inject all local stores and let the user mutate them directly
          const stores = mapStateToStores(state);
          super.deriveStoresFromProps(props, ...stores);
        }
        // respect user defined getDerivedStateFromProps
        if (super.getDerivedStateFromProps) {
          return super.getDerivedStateFromProps(props, state);
        }
        return null;
      }

      componentWillUnmount() {
        // call user defined componentWillUnmount
        if (super.componentWillUnmount) {
          super.componentWillUnmount();
        }
        // clean up memory used by Easy State
        unobserve(this.render);
      }
    }

    ReactiveComp = ReactiveClassComp;
  }

  ReactiveComp.displayName = Comp.displayName || Comp.name;
  // static props are inherited by class components,
  // but have to be copied for function components
  if (isStatelessComp) {
    for (let key of Object.keys(Comp)) {
      ReactiveComp[key] = Comp[key];
    }
  }

  return ReactiveComp;
}

function mapStateToStores(state) {
  // find store properties and map them to their none observable raw value
  // to do not trigger none static this.setState calls
  // from the static getDerivedStateFromProps lifecycle method
  const component = state[COMPONENT];
  return Object.keys(component).map(key => component[key]).filter(isObservable).map(raw);
}

function store(obj) {
  // do not create new versions of the store on every render
  // if it is a local store in a function component
  // create a memoized store at the first call instead
  if (hasHooks && isInsideFunctionComponent) {
    return useMemo(() => observable(obj), []);
  }
  return observable(obj);
}

export { view, store, batch };
