(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('react')) :
	typeof define === 'function' && define.amd ? define(['exports', 'react'], factory) :
	(factory((global.easyState = {}),global.react));
}(this, (function (exports,react) { 'use strict';

var connectionStore = new WeakMap();
var ITERATION_KEY = Symbol('iteration key');

function storeObservable(obj) {
  // this will be used to save (obj.key -> reaction) connections later
  connectionStore.set(obj, Object.create(null));
}

function registerReactionForOperation(reaction, ref) {
  var target = ref.target;
  var key = ref.key;
  var type = ref.type;

  if (type === 'iterate') {
    key = ITERATION_KEY;
  }

  var reactionsForObj = connectionStore.get(target);
  var reactionsForKey = reactionsForObj[key];
  if (!reactionsForKey) {
    reactionsForObj[key] = reactionsForKey = new Set();
  }
  // save the fact that the key is used by the reaction during its current run
  if (!reactionsForKey.has(reaction)) {
    reactionsForKey.add(reaction);
    reaction.cleaners.push(reactionsForKey);
  }
}

function getReactionsForOperation(ref) {
  var target = ref.target;
  var key = ref.key;
  var type = ref.type;

  var reactionsForTarget = connectionStore.get(target);
  var reactionsForKey = new Set();

  if (type === 'clear') {
    for (var key$1 in reactionsForTarget) {
      addReactionsForKey(reactionsForKey, reactionsForTarget, key$1);
    }
  } else {
    addReactionsForKey(reactionsForKey, reactionsForTarget, key);
  }

  if (type === 'add' || type === 'delete' || type === 'clear') {
    var iterationKey = Array.isArray(target) ? 'length' : ITERATION_KEY;
    addReactionsForKey(reactionsForKey, reactionsForTarget, iterationKey);
  }

  return reactionsForKey;
}

function addReactionsForKey(reactionsForKey, reactionsForTarget, key) {
  var reactions = reactionsForTarget[key];
  reactions && reactions.forEach(reactionsForKey.add, reactionsForKey);
}

function releaseReaction(reaction) {
  if (reaction.cleaners) {
    reaction.cleaners.forEach(releaseReactionKeyConnection, reaction);
  }
  reaction.cleaners = [];
}

function releaseReactionKeyConnection(reactionsForKey) {
  reactionsForKey.delete(this);
}

var runningReaction;
var isDebugging = false;

function runAsReaction(reaction, fn, context, args) {
  // do not build reactive relations, if the reaction is unobserved
  if (reaction.unobserved) {
    return fn.apply(context, args);
  }

  // release the (obj -> key -> reactions) connections
  // and reset the cleaner connections
  releaseReaction(reaction);

  try {
    // set the reaction as the currently running one
    // this is required so that we can create (observable.prop -> reaction) pairs in the get trap
    runningReaction = reaction;
    return fn.apply(context, args);
  } finally {
    // always remove the currently running flag from the reaction when it stops execution
    runningReaction = undefined;
  }
}

// register the currently running reaction to be queued again on obj.key mutations
function registerRunningReactionForOperation(operation) {
  if (runningReaction) {
    debugOperation(runningReaction, operation);
    registerReactionForOperation(runningReaction, operation);
  }
}

function queueReactionsForOperation(operation) {
  // iterate and queue every reaction, which is triggered by obj.key mutation
  getReactionsForOperation(operation).forEach(queueReaction, operation);
}

function queueReaction(reaction) {
  debugOperation(reaction, this);
  // queue the reaction for later execution or run it immediately
  if (typeof reaction.scheduler === 'function') {
    reaction.scheduler(reaction);
  } else if (typeof reaction.scheduler === 'object') {
    reaction.scheduler.add(reaction);
  } else {
    reaction();
  }
}

function debugOperation(reaction, operation) {
  if (reaction.debugger && !isDebugging) {
    try {
      isDebugging = true;
      reaction.debugger(operation);
    } finally {
      isDebugging = false;
    }
  }
}

function hasRunningReaction() {
  return runningReaction !== undefined;
}

var IS_REACTION = Symbol('is reaction');

function observe(fn, options) {
  if ( options === void 0 ) { options = {}; }

  // wrap the passed function in a reaction, if it is not already one
  var reaction = fn[IS_REACTION] ? fn : function reaction() {
    return runAsReaction(reaction, fn, this, arguments);
  };
  // save the scheduler and debugger on the reaction
  reaction.scheduler = options.scheduler;
  reaction.debugger = options.debugger;
  // save the fact that this is a reaction
  reaction[IS_REACTION] = true;
  // run the reaction once if it is not a lazy one
  if (!options.lazy) {
    reaction();
  }
  return reaction;
}

function unobserve(reaction) {
  // do nothing, if the reaction is already unobserved
  if (!reaction.unobserved) {
    // indicate that the reaction should not be triggered any more
    reaction.unobserved = true;
    // release (obj -> key -> reaction) connections
    releaseReaction(reaction);
  }
  // unschedule the reaction, if it is scheduled
  if (typeof reaction.scheduler === 'object') {
    reaction.scheduler.delete(reaction);
  }
}

var proxyToRaw = new WeakMap();
var rawToProxy = new WeakMap();

var getPrototypeOf = Object.getPrototypeOf;
var hasOwnProperty = Object.prototype.hasOwnProperty;

var instrumentations = {
  has: function has(key) {
    var target = proxyToRaw.get(this);
    var proto = getPrototypeOf(this);
    registerRunningReactionForOperation({ target: target, key: key, type: 'has' });
    return proto.has.apply(target, arguments);
  },
  get: function get(key) {
    var target = proxyToRaw.get(this);
    var proto = getPrototypeOf(this);
    registerRunningReactionForOperation({ target: target, key: key, type: 'get' });
    return proto.get.apply(target, arguments);
  },
  add: function add(key) {
    var target = proxyToRaw.get(this);
    var proto = getPrototypeOf(this);
    var hadKey = proto.has.call(target, key);
    // forward the operation before queueing reactions
    var result = proto.add.apply(target, arguments);
    if (!hadKey) {
      queueReactionsForOperation({ target: target, key: key, value: key, type: 'add' });
    }
    return result;
  },
  set: function set(key, value) {
    var target = proxyToRaw.get(this);
    var proto = getPrototypeOf(this);
    var hadKey = proto.has.call(target, key);
    var oldValue = proto.get.call(target, key);
    // forward the operation before queueing reactions
    var result = proto.set.apply(target, arguments);
    if (!hadKey) {
      queueReactionsForOperation({ target: target, key: key, value: value, type: 'add' });
    } else if (value !== oldValue) {
      queueReactionsForOperation({ target: target, key: key, value: value, oldValue: oldValue, type: 'set' });
    }
    return result;
  },
  delete: function delete$1(key) {
    var target = proxyToRaw.get(this);
    var proto = getPrototypeOf(this);
    var hadKey = proto.has.call(target, key);
    var oldValue = proto.get ? proto.get.call(target, key) : undefined;
    // forward the operation before queueing reactions
    var result = proto.delete.apply(target, arguments);
    if (hadKey) {
      queueReactionsForOperation({ target: target, key: key, oldValue: oldValue, type: 'delete' });
    }
    return result;
  },
  clear: function clear() {
    var target = proxyToRaw.get(this);
    var proto = getPrototypeOf(this);
    var hadItems = target.size !== 0;
    var oldTarget = target instanceof Map ? new Map(target) : new Set(target);
    // forward the operation before queueing reactions
    var result = proto.clear.apply(target, arguments);
    if (hadItems) {
      queueReactionsForOperation({ target: target, oldTarget: oldTarget, type: 'clear' });
    }
    return result;
  },
  forEach: function forEach() {
    var target = proxyToRaw.get(this);
    var proto = getPrototypeOf(this);
    registerRunningReactionForOperation({ target: target, type: 'iterate' });
    return proto.forEach.apply(target, arguments);
  },
  keys: function keys() {
    var target = proxyToRaw.get(this);
    var proto = getPrototypeOf(this);
    registerRunningReactionForOperation({ target: target, type: 'iterate' });
    return proto.keys.apply(target, arguments);
  },
  values: function values() {
    var target = proxyToRaw.get(this);
    var proto = getPrototypeOf(this);
    registerRunningReactionForOperation({ target: target, type: 'iterate' });
    return proto.values.apply(target, arguments);
  },
  entries: function entries() {
    var target = proxyToRaw.get(this);
    var proto = getPrototypeOf(this);
    registerRunningReactionForOperation({ target: target, type: 'iterate' });
    return proto.entries.apply(target, arguments);
  },
  get size() {
    var target = proxyToRaw.get(this);
    var proto = getPrototypeOf(this);
    registerRunningReactionForOperation({ target: target, type: 'iterate' });
    return Reflect.get(proto, 'size', target);
  }
};
instrumentations[Symbol.iterator] = function () {
    var target = proxyToRaw.get(this);
    var proto = getPrototypeOf(this);
    registerRunningReactionForOperation({ target: target, type: 'iterate' });
    return proto[Symbol.iterator].apply(target, arguments);
  };

var collectionHandlers = {
  get: function get(target, key, receiver) {
    // instrument methods and property accessors to be reactive
    target = hasOwnProperty.call(instrumentations, key) ? instrumentations : target;
    return Reflect.get(target, key, receiver);
  }
};

// eslint-disable-next-line
var globalObj = Function("return this")();

// built-in object can not be wrapped by Proxies
// their methods expect the object instance as the 'this' instead of the Proxy wrapper
// complex objects are wrapped with a Proxy of instrumented methods
// which switch the proxy to the raw object and to add reactive wiring
var handlers = new Map([[Map, collectionHandlers], [Set, collectionHandlers], [WeakMap, collectionHandlers], [WeakSet, collectionHandlers], [Object, false], [Array, false], [Int8Array, false], [Uint8Array, false], [Uint8ClampedArray, false], [Int16Array, false], [Uint16Array, false], [Int32Array, false], [Uint32Array, false], [Float32Array, false], [Float64Array, false]]);

function shouldInstrument(ref) {
  var constructor = ref.constructor;

  var isBuiltIn = typeof constructor === 'function' && constructor.name in globalObj && globalObj[constructor.name] === constructor;
  return !isBuiltIn || handlers.has(constructor);
}

function getHandlers(obj) {
  return handlers.get(obj.constructor);
}

var hasOwnProperty$1 = Object.prototype.hasOwnProperty;

// intercept get operations on observables to know which reaction uses their properties
function get(target, key, receiver) {
  var result = Reflect.get(target, key, receiver);
  // do not register (observable.prop -> reaction) pairs for these cases
  if (typeof key === 'symbol' || typeof result === 'function') {
    return result;
  }
  // register and save (observable.prop -> runningReaction)
  registerRunningReactionForOperation({ target: target, key: key, receiver: receiver, type: 'get' });
  // if we are inside a reaction and observable.prop is an object wrap it in an observable too
  // this is needed to intercept property access on that object too (dynamic observable tree)
  var observableResult = rawToProxy.get(result);
  if (hasRunningReaction() && typeof result === 'object' && result !== null) {
    if (observableResult) {
      return observableResult;
    }
    // do not violate the none-configurable none-writable prop get handler invariant
    // fall back to none reactive mode in this case, instead of letting the Proxy throw a TypeError
    var descriptor = Reflect.getOwnPropertyDescriptor(target, key);
    if (!descriptor || !(descriptor.writable === false && descriptor.configurable === false)) {
      return observable(result);
    }
  }
  // otherwise return the observable wrapper if it is already created and cached or the raw object
  return observableResult || result;
}

function has(target, key) {
  var result = Reflect.has(target, key);
  // do not register (observable.prop -> reaction) pairs for these cases
  if (typeof key === 'symbol') {
    return result;
  }
  // register and save (observable.prop -> runningReaction)
  registerRunningReactionForOperation({ target: target, key: key, type: 'has' });
  return result;
}

function ownKeys(target) {
  registerRunningReactionForOperation({ target: target, type: 'iterate' });
  return Reflect.ownKeys(target);
}

// intercept set operations on observables to know when to trigger reactions
function set(target, key, value, receiver) {
  // make sure to do not pollute the raw object with observables
  if (typeof value === 'object' && value !== null) {
    value = proxyToRaw.get(value) || value;
  }
  // save if the object had a descriptor for this key
  var hadKey = hasOwnProperty$1.call(target, key);
  // save if the value changed because of this set operation
  var oldValue = target[key];
  // execute the set operation before running any reaction
  var result = Reflect.set(target, key, value, receiver);
  // emit a warning and do not queue anything when another reaction is queued
  // from an already running reaction
  if (hasRunningReaction()) {
    console.error(("Mutating observables in reactions is forbidden. You set " + key + " to " + value + "."));
    return result;
  }
  // do not queue reactions if it is a symbol keyed property
  // or the target of the operation is not the raw receiver
  // (possible because of prototypal inheritance)
  if (typeof key === 'symbol' || target !== proxyToRaw.get(receiver)) {
    return result;
  }

  // queue a reaction if it's a new property or its value changed
  if (!hadKey) {
    queueReactionsForOperation({ target: target, key: key, value: value, receiver: receiver, type: 'add' });
  } else if (value !== oldValue) {
    queueReactionsForOperation({
      target: target,
      key: key,
      value: value,
      oldValue: oldValue,
      receiver: receiver,
      type: 'set'
    });
  }
  return result;
}

function deleteProperty(target, key) {
  // save if the object had the key
  var hadKey = hasOwnProperty$1.call(target, key);
  var oldValue = target[key];
  // execute the delete operation before running any reaction
  var result = Reflect.deleteProperty(target, key);
  // only queue reactions for non symbol keyed property delete which resulted in an actual change
  if (typeof key !== 'symbol' && hadKey) {
    queueReactionsForOperation({ target: target, key: key, oldValue: oldValue, type: 'delete' });
  }
  return result;
}

var baseHandlers = { get: get, has: has, ownKeys: ownKeys, set: set, deleteProperty: deleteProperty };

function observable(obj) {
  if ( obj === void 0 ) { obj = {}; }

  // if it is already an observable or it should not be wrapped, return it
  if (proxyToRaw.has(obj) || !shouldInstrument(obj)) {
    return obj;
  }
  // if it already has a cached observable wrapper, return it
  // otherwise create a new observable
  return rawToProxy.get(obj) || createObservable(obj);
}

function createObservable(obj) {
  // if it is a complex built-in object or a normal object, wrap it
  var handlers = getHandlers(obj) || baseHandlers;
  var observable = new Proxy(obj, handlers);
  // save these to switch between the raw object and the wrapped object with ease later
  rawToProxy.set(obj, observable);
  proxyToRaw.set(observable, obj);
  // init basic data structures to save and cleanup later (observable.prop -> reaction) connections
  storeObservable(obj);
  return observable;
}

function isObservable(obj) {
  return proxyToRaw.has(obj);
}

function raw(obj) {
  return proxyToRaw.get(obj) || obj;
}

var tasks = new Set();
var batchCount = 0;

function add(task) {
  if (batchCount !== 0) {
    tasks.add(task);
  } else {
    runTask(task);
  }
}

function remove(task) {
  tasks.delete(task);
}

// must remove the task before executing it
// toherwise executed task caould start a new task list execution
// and recursively run itself infinite times
function runTask(task) {
  remove(task);
  task();
}

// this runs the passed function and delays all re-renders
// until the function is finished running
function batch(fn, args) {
  try {
    batchCount++;
    return fn.apply(this, args);
  } finally {
    batchCount--;
    if (batchCount === 0) {
      tasks.forEach(runTask);
    }
  }
}

// this creates and returns a batched version of the passed function
// the cache is necessary to always map the same thing to the same function
// which makes sure that addEventListener/removeEventListener pairs don't break
var cache = new Map();
function batchFn(fn) {
  if (typeof fn !== 'function') {
    return fn;
  }
  var batched = cache.get(fn);
  if (!batched) {
    batched = function () {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      return batch(fn, args);
    };
    cache.set(fn, batched);
  }
  return batched;
}

// batched window.addEventListener(cb) like callbacks
function batchCallbacks(functionWithCallbacks) {
  return function batchedCallbacks() {
    var args = [], len = arguments.length;
    while ( len-- ) args[ len ] = arguments[ len ];

    return functionWithCallbacks.apply(this, args.map(batchFn));
  };
}

// batches obj.onevent = fn like calls
function batchMethod(obj, method) {
  var descriptor = Object.getOwnPropertyDescriptor(obj, method);
  if (descriptor) {
    var newDescriptor = Object.assign({}, descriptor, {
      set: function set(value) {
        return descriptor.set.call(this, batchFn(value));
      }
    });
    Object.defineProperty(obj, method, newDescriptor);
  }
}

// try to find the global object
// it is window in the DOM and global in NodeJS and React Native
var globalObj$1;
if (typeof window !== 'undefined') {
  globalObj$1 = window;
} else if (typeof global !== 'undefined') {
  globalObj$1 = global;
}

// do a sync batching for the most common task sources
// this should be removed when React's own batching is improved in the future
if (globalObj$1) {
  globalObj$1.setTimeout = batchCallbacks(globalObj$1.setTimeout);
  globalObj$1.setInterval = batchCallbacks(globalObj$1.setInterval);
  if (globalObj$1.requestAnimationFrame) {
    globalObj$1.requestAnimationFrame = batchCallbacks(globalObj$1.requestAnimationFrame);
  }
  if (globalObj$1.requestIdleCallback) {
    globalObj$1.requestIdleCallback = batchCallbacks(globalObj$1.requestIdleCallback);
  }
  // eslint-disable-next-line
  Promise.prototype.then = batchCallbacks(Promise.prototype.then);
  // eslint-disable-next-line
  Promise.prototype.catch = batchCallbacks(Promise.prototype.catch);

  // this batches websocket event handlers
  if (globalObj$1.WebSocket) {
    var websocketHandlers = ['onopen', 'onmessage', 'onerror', 'onclose'];
    websocketHandlers.forEach(function (method) { return batchMethod(globalObj$1.WebSocket.prototype, method); });
  }
}

// batch addEventListener calls
if (globalObj$1.EventTarget) {
  EventTarget.prototype.addEventListener = batchCallbacks(EventTarget.prototype.addEventListener);
  EventTarget.prototype.removeEventListener = batchCallbacks(EventTarget.prototype.removeEventListener);
}

// HTTP event handlers are usually wrapped by Promises, which is covered above

var hasHooks = typeof react.useState === 'function';

var isInsideFunctionComponent = false;
var COMPONENT = Symbol('owner component');
var DUMMY_STATE = {};

function view(Comp) {
  var isStatelessComp = !(Comp.prototype && Comp.prototype.isReactComponent);

  var ReactiveComp;

  if (isStatelessComp && hasHooks) {
    // use a hook based reactive wrapper when we can
    ReactiveComp = react.memo(function (props) {
      var ref = react.useState();
      var setState = ref[1];

      // create a memoized reactive wrapper of the original component (render)
      // at the very first run of the component function
      var render = react.useMemo(function () {
        // run a dummy setState to schedule a new render
        var updater = function () { return setState(DUMMY_STATE); };

        return observe(Comp, {
          scheduler: {
            add: function () { return add(updater); },
            delete: function () { return remove(updater); }
          },
          lazy: true
        });
      }, []);

      // cleanup the reactive connections after the very last render of the component
      react.useEffect(function () {
        return function () { return unobserve(render); };
      }, []);

      // the isInsideFunctionComponent flag is used to toggle `store` behavior
      // based on where it was called from
      isInsideFunctionComponent = true;
      try {
        // run the reactive render instead of the original one
        return render(props);
      } finally {
        isInsideFunctionComponent = false;
      }
    });
  } else {
    var BaseComp = isStatelessComp ? react.Component : Comp;
    // a HOC which overwrites render, shouldComponentUpdate and componentWillUnmount
    // it decides when to run the new reactive methods and when to proxy to the original methods
    var ReactiveClassComp = (function (BaseComp) {
      function ReactiveClassComp(props, context) {
        var this$1 = this;

        BaseComp.call(this, props, context);

        this.state = this.state || {};
        this.state[COMPONENT] = this;

        // run a dummy setState to schedule a new render, avoid forceUpdate
        var updater = function () { return this$1.setState(DUMMY_STATE); };

        // create a reactive render for the component
        this.render = observe(this.render, {
          scheduler: {
            add: function () { return add(updater); },
            delete: function () { return remove(updater); }
          },
          lazy: true
        });
      }

      if ( BaseComp ) ReactiveClassComp.__proto__ = BaseComp;
      ReactiveClassComp.prototype = Object.create( BaseComp && BaseComp.prototype );
      ReactiveClassComp.prototype.constructor = ReactiveClassComp;

      ReactiveClassComp.prototype.render = function render () {
        return isStatelessComp ? Comp(this.props, this.context) : BaseComp.prototype.render.call(this);
      };

      // react should trigger updates on prop changes, while easyState handles store changes
      ReactiveClassComp.prototype.shouldComponentUpdate = function shouldComponentUpdate (nextProps, nextState) {
        var ref = this;
        var props = ref.props;
        var state = ref.state;

        // respect the case when user prohibits updates
        if (BaseComp.prototype.shouldComponentUpdate && !BaseComp.prototype.shouldComponentUpdate.call(this, nextProps, nextState)) {
          return false;
        }

        // return true if it is a reactive render or state changes
        if (state !== nextState) {
          return true;
        }

        // the component should update if any of its props shallowly changed value
        var keys = Object.keys(props);
        var nextKeys = Object.keys(nextProps);
        return nextKeys.length !== keys.length || nextKeys.some(function (key) { return props[key] !== nextProps[key]; });
      };

      // add a custom deriveStoresFromProps lifecyle method
      ReactiveClassComp.getDerivedStateFromProps = function getDerivedStateFromProps (props, state) {
        if (BaseComp.deriveStoresFromProps) {
          // inject all local stores and let the user mutate them directly
          var stores = mapStateToStores(state);
          BaseComp.deriveStoresFromProps.apply(this, [ props ].concat( stores ));
        }
        // respect user defined getDerivedStateFromProps
        if (BaseComp.getDerivedStateFromProps) {
          return BaseComp.getDerivedStateFromProps.call(this, props, state);
        }
        return null;
      };

      ReactiveClassComp.prototype.componentWillUnmount = function componentWillUnmount () {
        // call user defined componentWillUnmount
        if (BaseComp.prototype.componentWillUnmount) {
          BaseComp.prototype.componentWillUnmount.call(this);
        }
        // clean up memory used by Easy State
        unobserve(this.render);
      };

      return ReactiveClassComp;
    }(BaseComp));

    ReactiveComp = ReactiveClassComp;
  }

  ReactiveComp.displayName = Comp.displayName || Comp.name;
  // static props are inherited by class components,
  // but have to be copied for function components
  if (isStatelessComp) {
    for (var i = 0, list = Object.keys(Comp); i < list.length; i += 1) {
      var key = list[i];

      ReactiveComp[key] = Comp[key];
    }
  }

  return ReactiveComp;
}

function mapStateToStores(state) {
  // find store properties and map them to their none observable raw value
  // to do not trigger none static this.setState calls
  // from the static getDerivedStateFromProps lifecycle method
  var component = state[COMPONENT];
  return Object.keys(component).map(function (key) { return component[key]; }).filter(isObservable).map(raw);
}

function store(obj) {
  // do not create new versions of the store on every render
  // if it is a local store in a function component
  // create a memoized store at the first call instead
  if (hasHooks && isInsideFunctionComponent) {
    return react.useMemo(function () { return observable(obj); }, []);
  }
  return observable(obj);
}

exports.view = view;
exports.store = store;
exports.batch = batch;

Object.defineProperty(exports, '__esModule', { value: true });

})));
